package com.vanheusden.pfa;

public final class Move {
	/* private stuff */
	private final int fx, fy, tx, ty;
	private final MoveType moveType;
	private final ChessObjectType promoteTo;
	private Scene scene;

	/* public stuff */
	public Move(final int fx, final int fy, final int tx, final int ty, final MoveType moveType, final Scene scene, final ChessObjectType promoteTo) {
		this.fx = fx;
		this.fy = fy;
		this.tx = tx;
		this.ty = ty;
		this.moveType = moveType;
		this.scene = scene;
		this.promoteTo = promoteTo;
	}

	public Move(final int fx, final int fy, final int tx, final int ty, final ChessObjectType promoteTo) {
		this.fx = fx;
		this.fy = fy;
		this.tx = tx;
		this.ty = ty;
		this.moveType = MoveType.REGULAR;
		this.promoteTo = promoteTo;
	}

	public Move(final int fx, final int fy, final int tx, final int ty, final MoveType moveType) {
		this.fx = fx;
		this.fy = fy;
		this.tx = tx;
		this.ty = ty;
		this.moveType = moveType;
		this.promoteTo = null;
	}

	public Move(final int fx, final int fy, final int tx, final int ty) {
		this.fx = fx;
		this.fy = fy;
		this.tx = tx;
		this.ty = ty;
		this.moveType = MoveType.REGULAR;
		this.promoteTo = null;
	}

	public Move(final int fx, final int fy, final int tx) {
		this.fx = fx;
		this.fy = fy;
		this.tx = tx;
		this.ty = fy;
		this.moveType = MoveType.CASTLING;
		this.promoteTo = null;
	}

	public Move(byte [] bytes) {
		fx = bytes[0] & 0x0f;
		fy = (bytes[0] >> 4) & 0x0f;
		tx = bytes[1] & 0x0f;
		ty = (bytes[1] >> 4) & 0x0f;
		moveType = MoveType.getMT((byte)(bytes[2] & 0x0f));
		promoteTo = ChessObjectType.getCOT((byte)((bytes[2] >> 4) & 0x0f));
	}

	public Move duplicate() {
		return new Move(fx, fy, tx, ty, moveType, scene, promoteTo);
	}

	public void toBytes(byte [] out) {
		byte mt = moveType == null ? MoveType.getNullType() : moveType.getType();
		byte pt = promoteTo == null ? ChessObjectType.getSVINullType() : promoteTo.getSVI();

		out[0] = (byte)(fx | (fy << 4));
		out[1] = (byte)(tx | (ty << 4));
		out[2] = (byte)(mt | (pt << 4));
	}

	public String toString() {
		return toString(null);
	}

	public String toString(Board b) {
		String from = "", to = "";

		if (b != null) {
			ChessObject src = b.getAt(fx, fy);

			from = src.getType().getName();

			ChessObject dest = b.getAt(tx, ty);

			if (dest != null)
				to += dest.getType().getName();
		}

		from += "" + (char)('a' + fx) + (char)('1' + fy);

		to += "" + (char)('a' + tx) + (char)('1' + ty);

		if (b != null)
			return from + "-" + to;

		return from + to; // FIXME promotions
	}

	public boolean isA1A1() {
		return fx == 0 && fy == 0 && tx == 0 && ty == 0;
	}

	public int getFX() {
		return fx;
	}

	public int getFY() {
		return fy;
	}

	public int getTX() {
		return tx;
	}

	public int getTY() {
		return ty;
	}

	public MoveType getMoveType() {
		return moveType;
	}

	public ChessObjectType getPromoteTo() {
		return promoteTo;
	}

	public Scene getScene() {
		return scene;
	}

	public void setScene(final Scene scene) {
		assert this.scene == null || scene == null;
		this.scene = scene;
	}

	// QUICK & DIRTY!
	public boolean compareTo(final String move) {
		final String mlc = move.toLowerCase();

		final int ifx = mlc.charAt(0) - 'a';
		final int ify = mlc.charAt(1) - '1';

		final int to = mlc.charAt(2) == '-' ? 3 : 2;

		final int itx = mlc.charAt(to + 0) - 'a';
		final int ity = mlc.charAt(to + 1) - '1';

		return ifx == fx &&
			ify == fy &&
			itx == tx &&
			ity == ty;
	}

	@Override
	public boolean equals(final Object moveIn) {
		Move move = (Move)moveIn;

		return move.getFX() == fx &&
			move.getFY() == fy &&
			move.getTX() == tx &&
			move.getTY() == ty &&
			move.getMoveType() == moveType &&
			move.getPromoteTo() == promoteTo;
	}

	public boolean compareTo(final Move move, final boolean ignoreScene, final boolean ignorePromoteTo) {
		boolean result = move.getFX() == fx &&
			         move.getFY() == fy &&
			         move.getTX() == tx &&
			         move.getTY() == ty &&
			         move.getMoveType() == moveType;

		if (ignoreScene == false) {
			if (move.getScene() != scene)
				result = false;
		}

		if (ignorePromoteTo == false) {
			if (move.getPromoteTo() != promoteTo)
				result = false;
		}

		return result;
	}
}
