package com.vanheusden.pfa;

import java.io.*;
import java.lang.Thread.State;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.*;
import java.util.concurrent.locks.*;

class Perft {
	int maxThreadsPerSlot = 0, runningThreadsCount = 0;
	final Semaphore threadCountLock = new Semaphore(1, false);
	final boolean verbose;

	Perft(final boolean verbose) {
		this.verbose = verbose;
	}

	///// generic methods /////
	Scene getSceneAfterMove(Scene scene, Move currentMove, PlayerColor side) {
		Scene sceneAfterMove = currentMove.getScene();

		if (sceneAfterMove == null)
			sceneAfterMove = scene.Move(currentMove);
		else
			currentMove.setScene(null); // help GC, link is not required

		sceneAfterMove.validateMoves(side);

		return sceneAfterMove;
	}

	///// methods for single threaded perft /////
	public long searchST(final Scene scene, final int currentDepth, final PlayerColor side) {
		scene.validateMoves(side);

		final List<Move> moves = scene.getMoveList(side);
		int nMoves = moves.size();

		int nextDepth = currentDepth - 1;

		if (nextDepth == 0)
			return nMoves;

		PlayerColor otherSide = Statics.opponentColor(side);

		long nodes = 0;

		for(int moveIndex=0; moveIndex < nMoves; moveIndex++) {
			final Move currentMove = moves.get(moveIndex);

			final Scene sceneAfterMove = getSceneAfterMove(scene, currentMove, otherSide);

			nodes += searchST(sceneAfterMove, nextDepth, otherSide);
		}

		return nodes;
	}

	///// methods for multi threaded perft /////
	class BrainThread implements Runnable {
		private final Scene scene;
		private final int currentDepth;
		private final PlayerColor side;
		private final Move currentMove;
		private Thread t;
		private long nodes;

		BrainThread(final Scene scene, final Move currentMove, final int currentDepth, final PlayerColor side) {
			this.scene = scene;
			this.currentDepth = currentDepth;
			this.side = side;
			this.currentMove = currentMove;
		}

		public void setThread(final Thread t) {
			this.t = t;
		}

		public Thread getThread() {
			return t;
		}

		public long getNodeCount() {
			return nodes;
		}

		public void run() {
			final Scene sceneAfterMove = getSceneAfterMove(scene, currentMove, side);

			try {
				nodes = searchMT(sceneAfterMove, currentDepth, side, true);
			}
			catch(InterruptedException ie) {
			}
		}
	}

	void endThreadAdminUpdate() {
		threadCountLock.acquireUninterruptibly();

		runningThreadsCount--;

		threadCountLock.release();
	}

	long processThreadsResults(final List<BrainThread> btList) throws InterruptedException {
		long nodes = 0;

		for(int index=0; index<btList.size();) {
			final BrainThread bt = btList.get(index);
			final Thread t = bt.getThread();

			if (t.getState() == State.TERMINATED) {
				t.join();

				btList.remove(index);

				nodes += bt.getNodeCount();
			}
			else {
				index++;
			}
		}

		return nodes;
	}

	public long searchMT(final Scene scene, final int currentDepth, final PlayerColor side, final boolean isThread) throws InterruptedException {
		scene.validateMoves(side);

		final List<Move> moves = scene.getMoveList(side);
		int nMoves = moves.size();

		int nextDepth = currentDepth - 1;

		if (nextDepth == 0) {
			if (isThread)
				endThreadAdminUpdate();

			return nMoves;
		}

		final List<BrainThread> btList = new ArrayList<BrainThread>();

		PlayerColor otherSide = Statics.opponentColor(side);

		long nodes = 0;

		boolean threadsStarted = false;

		for(int moveIndex=0; moveIndex < nMoves;) {
			if (currentDepth >= 1 && threadCountLock.tryAcquire()) {
				final int nToAdd = Math.min(nMoves - moveIndex, maxThreadsPerSlot - runningThreadsCount);
				assert(nToAdd >= 0);

				runningThreadsCount += nToAdd;

				threadCountLock.release();

				if (nToAdd > 0) {
					threadsStarted = true;

					for(int newThreadNr=0; newThreadNr < nToAdd; newThreadNr++) {
						final BrainThread bt = new BrainThread(scene, moves.get(moveIndex), nextDepth, otherSide);

						final Thread t = new Thread(bt);
						t.start();

						bt.setThread(t);
						btList.add(bt);

						moveIndex++;
					}
				}
			}

			if (moveIndex < nMoves) {
				final Move currentMove = moves.get(moveIndex++);

				final Scene sceneAfterMove = getSceneAfterMove(scene, currentMove, otherSide);

				nodes += searchMT(sceneAfterMove, nextDepth, otherSide, false);
			}

			if (threadsStarted)
				nodes += processThreadsResults(btList);
		}

		// we decrease the count here as the following loop will not (hardly) use cpu
		if (isThread)
			endThreadAdminUpdate();

		// thread results
		for(;;) {
			threadCountLock.acquireUninterruptibly();
			int nt = runningThreadsCount;
			threadCountLock.release();

			nodes += processThreadsResults(btList);

			if (btList.size() == 0)
				break;

			btList.get(0).getThread().join();
		}

		return nodes;
	}

	class RunStats implements Runnable {
		long prevTs = -1, prevNodeCount = -1;
		double maxNPS = -1.0;

		public void run() {
			try {
				Thread.sleep(1000);

				final Thread me = Thread.currentThread();

				// count running threads
				final ThreadGroup tg = me.getThreadGroup();

				final Thread [] list = new Thread[1024];

				for(;!Thread.interrupted();) {
					final int n = tg.enumerate(list);

					final long myId = me.getId();

					int cntRunnable = 0, cntBlocked = 0;
					for(int i=0; i<n; i++) {
						final Thread.State st = list[i].getState();

						if (st == State.RUNNABLE) {
							final long curId = list[i].getId();

							if (curId != myId) {
								cntRunnable++;

/*
								StackTraceElement [] sts = list[i].getStackTrace();
								StackTraceElement ste = sts[0];

								io.progressCallbackDebug(" " + list[i].getName() + "] " + ste.getClassName() + ", " + ste.getFileName() + ", " + ste.getLineNumber() + ", " + ste.getMethodName() + ", " + (ste.isNativeMethod() ?  "is" : "NOT a") + " native method");
*/
							}
						}
						else if (st != State.NEW && st != State.TERMINATED) {
							final long curId = list[i].getId();

							if (curId != myId) {
								cntBlocked++;
/*
								final StackTraceElement [] sts = list[i].getStackTrace();
								for(int nr=0; nr<Math.min(sts.length, 3); nr++) {
									StackTraceElement ste = sts[nr];

									io.progressCallbackDebug(" " + nr + "] " + list[i].getName() + "] " + ste.getClassName() + ", " + ste.getFileName() + ", " + ste.getLineNumber() + ", " + ste.getMethodName() + ", " + (ste.isNativeMethod() ?  "is" : "NOT a") + " native method");
								}
*/
							}
						}
					}

					String out = "runnable: " + cntRunnable + ", blocked: " + cntBlocked;

					System.err.println(out);

					Thread.sleep(3000);
				}
			}
			catch(InterruptedException bla) {
			}
		}
	}

	public long perft(final Scene s, final PlayerColor currentPlayer, final int depth, int nThreads) {
		if (nThreads == 0)
			nThreads = Runtime.getRuntime().availableProcessors();

		if (nThreads > 1) {
			maxThreadsPerSlot = Math.min(nThreads, Runtime.getRuntime().availableProcessors());

			try {
				RunStats rs = new RunStats();
				Thread rst = new Thread(rs, "RunStats");

				if (verbose)
					rst.start();

				long n = searchMT(s, depth, currentPlayer, false);

				rst.interrupt();

				return n;
			}
			catch(Exception e) {
				Statics.logException("perft", e);

				return -1;
			}
		}

		return searchST(s, depth, currentPlayer);
	}

	public long perft(final String epd, final int depth, final int nThreads) {
		PlayerColor currentPlayer = PlayerColor.WHITE;
		Scene s = null;

		if (epd == null) {
			s = new Scene();
		}
		else {
			Object [] obs = Statics.EPDToScene(epd);

			s = (Scene)obs[0];
			currentPlayer = (PlayerColor)obs[1];
		}

		return perft(s, currentPlayer, depth, nThreads);
	}
}
