package com.vanheusden.pfa;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.*;
import java.util.regex.*;

class XboardDrone extends Drone {
	AtomicBoolean setBoard = new AtomicBoolean(false);
	final int protover2MaxWait = 2500;
	Board initBoard = null;
	PlayerColor cur = PlayerColor.WHITE;
	Pattern thinkOutputPattern = Pattern.compile("^([0-9]+)[^ ]* +([0-9]+) +([0-9]+) +([0-9]+) +(.*)$");

	XboardDrone(final String host, final int port, final boolean canStop, final Map<String, String> parameters) throws IOException, InterruptedException {
		super(host, port, canStop, parameters);

		initBoard = new Scene().getBoard();

		pmq = new PMQuestioner();
	}

	XboardDrone(final String exe, final String path, final boolean canStop, final Map<String, String> parameters) throws IOException, InterruptedException {
		super(exe, path, canStop, parameters);

		initBoard = new Scene().getBoard();

		pmq = new PMQuestioner();
	}

	void initSession() throws IOException, InterruptedException {
		emit("xboard");
		emit("protover 2");

		String line = null;
		long initStart = System.currentTimeMillis();

		do {
			line = input.readLine();
			if (line == null)
				break;

			if (line.indexOf("setboard=1") != -1) {
				setBoard.set(true);

				logIgnoreIOException("\"setboard\" support enabled");
			}

			int nameIndex = line.indexOf("myname=");
			if (nameIndex != -1) {
				name = line.substring(nameIndex + 8);

				int endIndex = name.indexOf("\"");
				if (endIndex != -1)
					name = name.substring(0, endIndex);
			}

			if (getVerbose())
				logIgnoreIOException(getName() + " WF< " + line);
		}
		while(line.indexOf("done=1") == -1 && System.currentTimeMillis() - initStart < protover2MaxWait && !Thread.interrupted());

		if (line == null) {
			logIgnoreIOException(getName() + " no \"done=1\" from engine");

			disconnect();
			input = null;
		}

		emit("easy");
	}

	void stopPonder() {
		try {
			// do reset pondering; checkInput does not need to filter
			// anything like with uci
			if (pondering.getAndSet(false))
				emit("easy");
		}
		catch(IOException ie) {
			logIgnoreIOException(getName() + " IO Exception in stopPonder " + ie);

			disconnect();
		}
	}

	void startPonder(final Scene sc, final PlayerColor c, final Move lastMove) {
		try {
			if (!pondering.getAndSet(true))
				emit("hard");
		}
		catch(IOException ie) {
			logIgnoreIOException(getName() + " IO Exception in startPonder " + ie);

			disconnect();
		}
	}

	void doEditBoard(Board b) throws IOException {
		emit("edit");
		emit("#");

		PlayerColor cc = PlayerColor.WHITE;

		for(int y=0; y<8; y++) {
			for(int x=0; x<8; x++) {
				ChessObject co = b.getAt(x, y);
				if (co == null)
					continue;

				if (co.getColor() != cc) {
					emit("c");

					cc = Statics.opponentColor(cc);
				}

				String name = co.getType().getName().toUpperCase();
				String pos = "" + (char)('a' + x) + (char)('1' + y);

				emit(name + pos);
			}
		}

		emit(".");
	}

	public void startGame(final int nMoves, final long maxDuration) throws IOException, InterruptedException {
		emit("new");

		cur = PlayerColor.WHITE;

		if (nCores > 0)
			emit("cores " + nCores);

		long workMaxDuration = maxDuration / 1000; // make it seconds

		if (workMaxDuration < 0) {
			logIgnoreIOException(getName() + " startGame had a workduration of < 1 second");
			workMaxDuration = 1;
		}

		if (nMoves > 0)
			emit("level " + nMoves + " " + workMaxDuration / 60 + ":" + workMaxDuration % 60 + " 0");
		else
			emit("level 0 " + workMaxDuration / 60 + ":" + workMaxDuration % 60 + " 0");
	}

	class PMQuestioner extends Drone.PMQuestioner {
		void doAsk(PMQuestion q) {
			try {
				logIgnoreIOException(getName() + " executing question");

				initScene.acquireUninterruptibly();
				final Scene sc = scene;
				final PlayerColor c = color;
				final Move lastMove = q.getLastMove();
				final int movesLeft = nr;
				long useDeadLine = deadLine - latency;
				initScene.release();

				startAskTs = System.currentTimeMillis();

				long useTimeLeft = (useDeadLine - startAskTs) / 10;
				if (useTimeLeft < 1) {
					logIgnoreIOException(getName() + " no time left: skipping");
					return;
				}

				askedAnything.set(true);

				lastNps.set(-1);

				emit("force");

				if (setBoard.get()) {
					String cmdPosition = "setboard " + Statics.toFENString(sc.getBoard(), c, lastMove, 1, false);
					emit(cmdPosition);
				}
				else {
					if (lastMove != null)
						emit("" + lastMove);
				}

				emit("st 1");

				emit("time " + useTimeLeft);

				String cmdStart = "go";
				emit(cmdStart);

				cur = Statics.opponentColor(c);
			}
			catch(IOException ie) {
				logIgnoreIOException(getName() + " I/O exception in startAsk " + ie);

				disconnect();
			}
		}
	}

	public void interruptThinking() {
		// send sigint
		if (isYes(parameters.get("xboardsendstop"))) {
			try {
				emit("stop");
			}
			catch(IOException ie) {
				logIgnoreIOException(getName() + " IO Exception in stopAsk " + ie);

				disconnect();
			}
		}
	}

	public void stopAsk() {
		askedDuration = startAskTs = askReplyTs = -1;

		interruptThinking();
	}

	MoveResult checkInput(final String line) {
		String pollMoveString = null;

		if (line.length() > 12 && line.substring(0, 12).equals("My move is :"))
			pollMoveString = line.substring(12).trim();
		else if (line.length() > 4 && line.substring(0, 4).equals("move"))
			pollMoveString = line.substring(4).trim();

		if (pollMoveString != null) {
			askReplyTs = System.currentTimeMillis();

			StringBuffer err = new StringBuffer(""), err2 = new StringBuffer("");

                        initScene.acquireUninterruptibly();
                        Scene sc = scene;
                        PlayerColor c = color;
                        initScene.release();

			Move outTemp = Statics.LANToMove(pollMoveString, sc.getBoard(), c, sc.getMoveList(c), err);

			if (outTemp == null)
				outTemp = Statics.SANToMove(pollMoveString, sc.getBoard(), c, sc.getMoveList(c), err2);

			if (outTemp == null) {
				logIgnoreIOException(getName() + " illegal move " + pollMoveString + ": " + err.toString() + " / " + err2.toString());

				return null;
			}

			long took = askReplyTs - startAskTs;
			long percUsed = (took * 100) / askedDuration;
			logIgnoreIOException(getName() + " poll " + getName() + " choice: " + outTemp + ", took: " + took + "ms (" + percUsed + "%)");

			return new MoveResult(0, 0, outTemp, "drone:" + getName());
		}

		// FIXME handle statistics output (depth reached, score, pv etc)

		Matcher matcher = thinkOutputPattern.matcher(line);
		if (io != null && matcher.find()) {
			try {
				int ply = Integer.valueOf(matcher.group(1));
				int score = Integer.valueOf(matcher.group(2));
				int time = Integer.valueOf(matcher.group(3));
				long nodes = Long.valueOf(matcher.group(4));
				String pv = matcher.group(5);

				io.progressCallbackXboardShowThinking(ply, score, time, nodes, pv);
			}
			catch(Exception e) {
				Statics.logException("thinkoutput", e);
			}
		}

		return null;
	}
}
